/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be 
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified 
 *  version of the Library in object code or as an executable provided 
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete 
 *   corresponding machine-readable source code for the modified version 
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer 
 *   for a complete machine-readable copy of the corresponding source 
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#ifndef MARS_DEBUG_H
#define MARS_DEBUG_H

/*
 * \brief MARS Debug Macros
 */

#if defined(__cplusplus)
extern "C" {
#endif

#include <stdio.h>

#define MARS_DEBUG_ENABLE_PRINT		0
#define MARS_DEBUG_ENABLE_ASSERTS	0
#define MARS_DEBUG_ENABLE_VERBOSE	0
#define MARS_DEBUG_ENABLE_TIMER		0
#define MARS_DEBUG_ENABLE_DMA_ASSERTS	0
#define MARS_DEBUG_ENABLE_DMA_VERBOSE	0

#if __SPU__
#define __CONTEXT__		"MPU"
#define __BREAKPOINT__		__asm__ volatile ("stopd 0,1,1")
#else
#define __CONTEXT__		"HOST"
#define __BREAKPOINT__		__asm__ volatile ("tw 31,1,1")
#endif

#define __PRINT__(msg, ...) {\
	fprintf(stdout, "MARS("__CONTEXT__"): "msg, ##__VA_ARGS__);\
	fflush(stdout);\
}

#define MARS_PRINTF(msg, ...) {\
	__PRINT__("%s: "msg, __PRETTY_FUNCTION__, ##__VA_ARGS__);\
}

#if MARS_DEBUG_ENABLE_PRINT
#if __SPU__
#define MARS_PRINT(msg, ...) {\
	__PRINT__("KERNEL(%d): "msg, kernel_params.id, ##__VA_ARGS__);\
}
#else
#define MARS_PRINT(msg, ...) {\
	__PRINT__(msg, ##__VA_ARGS__);\
}
#endif

#define MARS_PRINT_16BYTE_ADDR(addr) {\
	int i;\
	printf("reg(%p) = 0x", (addr));\
	for (i = 0; i < 16; i++) {\
		printf("%02x", ((char *)(addr))[i]);\
	}\
	printf("\n");\
}

#define MARS_PRINT_REG(num) {\
	register volatile vector unsigned int *reg asm(#num);\
	MARS_PRINT_16BYTE_ADDR(reg);\
}

#define MARS_PRINT_STACK(stack_top, stack_pointer) {\
	char *sp = (char *)(stack_top);\
	while (sp >= (char *)(stack_pointer)) {\
		printf("[%p] ",sp);\
		int i;\
		for (i = 0; i < 4; i++)\
			printf("%02x",sp[i]);\
		printf("%s\n",sp==(char *)(stack_pointer)?" <-- $sp":"");\
		sp-=16;\
	} \
}
#else
#define MARS_PRINT(msg, ...)
#define MARS_PRINT_16BYTE_ADDR(addr)
#define MARS_PRINT_REG(num)
#define MARS_PRINT_STACK(stack_top, stack_pointer)
#endif

#if MARS_DEBUG_ENABLE_ASSERTS

#define MARS_BREAK() {\
	__PRINT__("%s[line %d]:%s():BREAK\n",\
		__FILE__, __LINE__, __PRETTY_FUNCTION__);\
	__BREAKPOINT__;\
}

#define MARS_ASSERT(exp) {\
	if(!(exp)) {\
		__PRINT__("%s(line %d):%s() - ASSERT("#exp") FAILED\n",\
			__FILE__, __LINE__, __PRETTY_FUNCTION__);\
		__BREAKPOINT__;\
	}\
}

#define MARS_CHECK(exp) {\
	MARS_ASSERT(exp);\
	if(!(exp)) return;\
}

#define MARS_CHECK_RET(exp, ret) {\
	MARS_ASSERT(exp);\
	if(!(exp)) return ret;\
}

#define MARS_CHECK_CLEANUP(exp, clean) {\
	MARS_ASSERT(exp);\
	if(!(exp)) {\
		clean;\
		return;\
	}\
}

#define MARS_CHECK_CLEANUP_RET(exp, clean, ret) {\
	MARS_ASSERT(exp);\
	if(!(exp)) {\
		clean;\
		return ret;\
	}\
}

#else
#define MARS_BREAK()
#define MARS_ASSERT(exp)
#define MARS_CHECK(exp)				{if(!(exp))return;}
#define MARS_CHECK_RET(exp, ret)		{if(!(exp))return ret;}
#define MARS_CHECK_CLEANUP(exp, clean)		{if(!(exp)){clean;return;}}
#define MARS_CHECK_CLEANUP_RET(exp, clean, ret)	{if(!(exp)){clean;return ret;}}
#endif

#if MARS_DEBUG_ENABLE_VERBOSE

#define MARS_PRINT_TASK_CONTEXT(task_context) {\
	__PRINT__("===\n");\
	__PRINT__("- Task Context (%p)\n", (task_context));\
	__PRINT__("-  exec:      %#llx\n", (task_context)->exec);\
	__PRINT__("-  exec_size: %d bytes\n", (task_context)->exec_size);\
	__PRINT__("-  bss_size:  %d bytes\n", (task_context)->bss_size);\
	__PRINT__("-  vaddr:     %#x\n", (task_context)->vaddr);\
	__PRINT__("-  entry:     %#x\n", (task_context)->entry);\
	__PRINT__("-  stack:     %#x\n", (task_context)->stack);\
	__PRINT__("===\n");\
}

#define MARS_PRINT_WORKLOAD_CONTEXT(workload_context) {\
	__PRINT__("===\n");\
	__PRINT__("- Workload Context (%p)\n", (workload_context));\
	__PRINT__("-  type:  %d\n",(workload_context)->type);\
	switch ((workload_context)->type) {\
	case MARS_WORKLOAD_TYPE_TASK:\
	MARS_PRINT_TASK_CONTEXT(&(workload_context)->type_context.task);\
	break;\
	}\
	__PRINT__("===\n");\
}

#define MARS_PRINT_WORKLOAD_QUEUE_HEADER(queue_header) {\
	__PRINT__("===\n");\
	__PRINT__("- Workload Queue Header (%p)\n", (queue_header));\
	__PRINT__("-  queue_ea:   %#llx\n",(queue_header)->queue_ea);\
	__PRINT__("-  context_ea: %#llx\n",(queue_header)->context_ea);\
	__PRINT__("-  flag:       %#x\n",(queue_header)->flag);\
	__PRINT__("-  count:      %d\n",(queue_header)->count);\
	unsigned int i;\
	__PRINT__("-  state:      ");\
	for (i = 0; i < MARS_WORKLOAD_MAX; i++)\
		fprintf(stdout, "%d,", (queue_header)->state[i]);\
	fprintf(stdout, "\n");\
	__PRINT__("-  priority:   ");\
	for (i = 0; i < MARS_WORKLOAD_MAX; i++)\
		fprintf(stdout, "%d,", (queue_header)->priority[i]);\
	fprintf(stdout, "\n");\
	__PRINT__("-  wait_count: ");\
	for (i = 0; i < MARS_WORKLOAD_MAX; i++)\
		fprintf(stdout, "%d,", (queue_header)->wait_count[i]);\
	fprintf(stdout, "\n");\
	__PRINT__("===\n");\
}

#define MARS_PRINT_KERNEL_PARAMS(kernel_params) {\
	__PRINT__("===\n");\
	__PRINT__("- Kernel Params (%p)\n", (kernel_params));\
	__PRINT__("-  workload_queue_ea: %#llx\n",\
			(kernel_params)->workload_queue_ea);\
	__PRINT__("===\n");\
}

#define MARS_PRINT_SPE_PROGRAM_HANDLE(spe_program_handle) {\
	__PRINT__("===\n");\
	__PRINT__("- SPE Program Handle (%p)\n", (spe_program_handle));\
	__PRINT__("-  handle_size: %d\n", (spe_program_handle)->handle_size);\
	__PRINT__("-  elf_image:   %p\n", (spe_program_handle)->elf_image);\
	__PRINT__("-  toe_shadow:  %p\n", (spe_program_handle)->toe_shadow);\
	__PRINT__("===\n");\
}

#define MARS_PRINT_ELF(elf) {\
	Elf32_Shdr *shdr = (Elf32_Shdr *)((char *)(elf) + ehdr->e_shoff);\
	char *str_table = (elf) + shdr[ehdr->e_shstrndx].sh_offset;\
	unsigned int i;\
	__PRINT__("===\n");\
	__PRINT__("- ELF Header (%p)\n", ehdr);\
	__PRINT__("-  e_type:      %#x\n", ehdr->e_type);\
	__PRINT__("-  e_machine:   %#x\n", ehdr->e_machine);\
	__PRINT__("-  e_version:   %#x\n", ehdr->e_version);\
	__PRINT__("-  e_entry:     %#x\n", ehdr->e_entry);\
	__PRINT__("-  e_phoff:     %#x\n", ehdr->e_phoff);\
	__PRINT__("-  e_shoff:     %#x\n", ehdr->e_shoff);\
	__PRINT__("-  e_flags:     %#x\n", ehdr->e_flags);\
	__PRINT__("-  e_ehsize:    %#x\n", ehdr->e_ehsize);\
	__PRINT__("-  e_phentsize: %#x\n", ehdr->e_phentsize);\
	__PRINT__("-  e_phnum:     %#x\n", ehdr->e_phnum);\
	__PRINT__("-  e_shentsize: %#x\n", ehdr->e_shentsize);\
	__PRINT__("-  e_shnum:     %#x\n", ehdr->e_shnum);\
	__PRINT__("-  e_shstrndx:  %#x\n", ehdr->e_shstrndx);\
	for (i = 0; i < ehdr->e_phnum; i++) {\
		__PRINT__("- ELF Program Header %d\n", i);\
		__PRINT__("-  p_type:       %#x\n", phdr[i].p_type);\
		__PRINT__("-  p_offset:     %#x\n", phdr[i].p_offset);\
		__PRINT__("-  p_vaddr:      %#x\n", phdr[i].p_vaddr);\
		__PRINT__("-  p_paddr:      %#x\n", phdr[i].p_paddr);\
		__PRINT__("-  p_filesz:     %d bytes\n", phdr[i].p_filesz);\
		__PRINT__("-  p_memsz:      %d bytes\n", phdr[i].p_memsz);\
		__PRINT__("-  p_flags:      %#x\n", phdr[i].p_flags);\
		__PRINT__("-  p_align:      %#x\n", phdr[i].p_align);\
	}\
	Elf32_Shdr *sh;\
	for (sh = shdr; sh < &shdr[ehdr->e_shnum]; ++sh) {\
		__PRINT__("- ELF Section Header %d\n", i);\
		__PRINT__("-  sh_name:      %s\n", str_table+sh->sh_name);\
	}\
	for (i = 0; i < ehdr->e_shnum; i++) {\
		__PRINT__("- ELF Section Header %d\n", i);\
		__PRINT__("-  sh_name:      %s\n", str_table+shdr[i].sh_name);\
		__PRINT__("-  sh_type:      %#x\n", shdr[i].sh_type);\
		__PRINT__("-  sh_flags:     %#x\n", shdr[i].sh_flags);\
		__PRINT__("-  sh_addr:      %#x\n", shdr[i].sh_addr);\
		__PRINT__("-  sh_offset:    %#x\n", shdr[i].sh_offset);\
		__PRINT__("-  sh_size:      %d bytes\n", shdr[i].sh_size);\
		__PRINT__("-  sh_link:      %#x\n", shdr[i].sh_link);\
		__PRINT__("-  sh_info:      %#x\n", shdr[i].sh_info);\
		__PRINT__("-  sh_addralign: %#x\n", shdr[i].sh_addralign);\
		__PRINT__("-  sh_entsize:   %d bytes\n", shdr[i].sh_entsize);\
	}\
	__PRINT__("===\n");\
}

#else
#define MARS_PRINT_TASK_CONTEXT(task_context)
#define MARS_PRINT_WORKLOAD_CONTEXT(workload_context)
#define MARS_PRINT_WORKLOAD_QUEUE_HEADER(queue_header)
#define MARS_PRINT_KERNEL_PARAMS(kernel_params)
#define MARS_PRINT_SPE_PROGRAM_HANDLE(spe_program_handle)
#define MARS_PRINT_ELF(elf)
#endif

#if MARS_DEBUG_ENABLE_TIMER
#define MARS_TIMER_START(tag)\
	{\
		struct mars_timer tag ## _timer;\
		mars_timer_start(&tag ## _timer);

#define MARS_TIMER_STOP(tag)\
		mars_timer_stop(&tag ## _timer);\
		__PRINT__("KERNEL(%d): " #tag " - %.0f us\n",\
			kernel_params.id, tag ## _timer.us);\
	}
#else
#define MARS_TIMER_START(id)
#define MARS_TIMER_STOP(id)
#endif

#if defined(__cplusplus)
}
#endif

#endif
