/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be 
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified 
 *  version of the Library in object code or as an executable provided 
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete 
 *   corresponding machine-readable source code for the modified version 
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer 
 *   for a complete machine-readable copy of the corresponding source 
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#ifndef MARS_TASK_TYPES_H
#define MARS_TASK_TYPES_H

/**
 * \file
 * \ingroup group_mars_task
 * \brief MARS Task Types
 */

#if defined(__cplusplus)
extern "C" {
#endif

#include <stdint.h>

/**
 * \ingroup group_mars_task
 * \brief Arguments structure size
 */
#define MARS_TASK_ARGS_SIZE			32

/**
 * \ingroup group_mars_task
 * \brief Arguments structure alignment
 */
#define MARS_TASK_ARGS_ALIGN			16

/**
 * \ingroup group_mars_task
 * \brief ID structure alignment
 */
#define MARS_TASK_ID_SIZE			32

/**
 * \ingroup group_mars_task
 * \brief ID structure alignment
 */
#define MARS_TASK_ID_ALIGN			16

/**
 * \ingroup group_mars_task
 * \brief Context structure size
 */
#define MARS_TASK_CONTEXT_SIZE			112

/**
 * \ingroup group_mars_task
 * \brief Context structure alignment
 */
#define MARS_TASK_CONTEXT_ALIGN			16

/**
 * \ingroup group_mars_task
 * \brief Context save area alignment
 */
#define MARS_TASK_CONTEXT_SAVE_ALIGN		128

/**
 * \ingroup group_mars_task
 * \brief Max context save area size
 * (LS size - task LS addr)
 */
#define MARS_TASK_CONTEXT_SAVE_SIZE_MAX		0x30000

/**
 * \ingroup group_mars_task
 * \brief Max length of task name
 */
#define MARS_TASK_NAME_LEN_MAX			22

/**
 * \ingroup group_mars_task
 * \brief MARS task argument structure
 *
 * This structure is initialized by the user and within the
 * \ref mars_task_params structure which is passed into
 * \ref mars_task_initialize for MARS task initialization.
 *
 * This argument structure is directly passed into the MARS
 * task's \ref mars_task_main function at task execution.
 */
struct mars_task_args {
	union {
		/** array of 32 8-bit unsigned ints */
		uint8_t  u8[32];
		/** array of 16 16-bit unsigned ints */
		uint16_t u16[16];
		/** array of 8 32-bit unsigned ints */
		uint32_t u32[8];
		/** array of 4 64-bit unsigned ints */
		uint64_t u64[4];
	} type;
} __attribute__((aligned(MARS_TASK_ARGS_ALIGN)));

/**
 * \ingroup group_mars_task
 * \brief MARS task id structure
 *
 * This structure is initialized during MARS task initialization and returned
 * when calling \ref mars_task_initialize.
 *
 * An instance of this structure must be kept until the task is finalized by
 * calling \ref mars_task_finalize.
 */
struct mars_task_id {
	uint64_t mars_context_ea;
	uint16_t workload_id;
	uint8_t	name[MARS_TASK_NAME_LEN_MAX];
} __attribute__((aligned(MARS_TASK_ID_ALIGN)));

struct mars_task_context {
	struct mars_task_id id;		/* task id */
	struct mars_task_args args;	/* task args */
	uint64_t exec;			/* ea of exec */
	uint32_t exec_size;		/* size of text and data of exec */
	uint32_t bss_size;		/* size of bss in memory of exec */
	uint32_t vaddr;			/* vaddr address of exec */
	uint32_t entry;			/* entry address of exec */
	uint32_t stack;			/* stack pointer of exec */
	uint32_t context_save_size;	/* context save size */
	uint64_t context_save_area;	/* context save area */
} __attribute__((aligned(MARS_TASK_CONTEXT_ALIGN)));

#if defined(__cplusplus)
}
#endif

#endif
