/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be 
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified 
 *  version of the Library in object code or as an executable provided 
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete 
 *   corresponding machine-readable source code for the modified version 
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer 
 *   for a complete machine-readable copy of the corresponding source 
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#ifndef MARS_TASK_H
#define MARS_TASK_H

/**
 * \file
 * \ingroup group_mars_task
 * \brief [host] MARS Task API
 */

#if defined(__cplusplus)
extern "C" {
#endif

#include "mars/mars_task_barrier.h"
#include "mars/mars_task_event_flag.h"
#include "mars/mars_task_queue.h"
#include "mars/mars_task_semaphore.h"
#include "mars/mars_task_signal.h"
#include "mars/mars_task_types.h"

struct mars_context;
struct mars_exec_context;

/**
 * \ingroup group_mars_task
 * \brief MARS task params structure
 *
 * This structure is initialized by the user and passed into
 * \ref mars_task_initialize for MARS task initialization.
 */
struct mars_task_params {
	/** name of task */
	const char *name;
	/** address of MPU program elf image */
	const void *elf_image;
	/** size of context save area allocated for task */
	uint32_t context_save_size;
};

/**
 * \ingroup group_mars_task
 * \brief [host] Initializes a MARS task.
 *
 * This function initializes a single task and adds it the MARS context's
 * workload queue.
 * Upon success, the specified id structure params will hold the initilized
 * task's id information.
 * You must call \ref mars_task_schedule in order for it to be scheduled for
 * execution by the kernel.
 * The task is in the finished state upon creation and may be finalized by
 * \ref mars_task_finalize without ever being scheduled for execution.
 *
 * When initializing a MARS task, you must specify its parameters through the
 * \ref mars_task_params structure:
 * \n
 * - \e name: The name is optional, but if specified to other than NULL its
 *	length must not exceed \ref MARS_TASK_NAME_LEN_MAX.
 * \n \n
 * - \e elf_image: The address of MPU program elf image in host storage that
 *	will be to be run by this task.
 * \n \n
 * - \e context_save_size: The size of context save area allocated and used
 *	during the time a task enters into the wait state. If 0 is specified,
 *	then no context save area will be allocated for the task and therefore
 *	the task must be a run-complete task. A run-complete task will run and
 *	occupy an MPU until it has completed running and exits. A run-complete
 *	task cannot enter the wait state, and therefore cannot call functions
 *	that will enter that task into a wait state. Currently either 0 or
 *	\ref MARS_TASK_CONTEXT_SAVE_SIZE_MAX must be specified.
 *
 * \param[in] mars		- pointer to initialized MARS context
 * \param[out] id		- pointer to task id to be initialized
 * \param[in] params		- pointer to task params of this task
 * \return
 *	MARS_SUCCESS		- successfully initialized MARS task
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_PARAMS	- bad task params specified
 * \n	MARS_ERROR_MEMORY	- not enough memory
 * \n	MARS_ERROR_LIMIT	- task queue is currently full
 */
int mars_task_initialize(struct mars_context *mars,
			struct mars_task_id *id,
			struct mars_task_params *params);

/**
 * \ingroup group_mars_task
 * \brief [host] Finalizes a MARS task.
 *
 * This function finalizes a task initialized by \ref mars_task_initialize.
 * The task will only be finalized if the task is in the finished state.
 * Once this function returns successfully and the task is finalized, the task
 * id should no longer be used.
 * If you want to make sure the task has finished before calling this function,
 * you can wait for task completion by calling \ref mars_task_wait or
 * \ref mars_task_try_wait.
 *
 * \param[in] id		- pointer to task id to finalize
 * \return
 *	MARS_SUCCESS		- successfully finalized MARS task
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_PARAMS	- bad task id specified
 * \n	MARS_ERROR_STATE	- task is in an invalid state
 */
int mars_task_finalize(struct mars_task_id *id);

/**
 * \ingroup group_mars_task
 * \brief [host] Schedules a MARS task for execution.
 *
 * This function schedules the task specified for execution.
 * The actual time of execution is determined by the scheduler.
 * Once the task is scheduled for execution by this function, it may not be
 * scheduled for execution until previous execution has finished.
 * You can wait for task completion by calling \ref mars_task_wait or
 * \ref mars_task_try_wait.
 *
 * - \e args: The task args (\ref mars_task_args) are optional, and only need
 *      to be passed in if the task executable expects it. If you specify NULL
 *      the task args passed into the task executable's main function will not
 *      be initialized.
 * \n \n
 * - \e priority: This is the priority of the task between 0 to 255, 0 being
 *      the lowest priority and 255 being the highest priority. Tasks with
 *      higher priority will be prioritized during scheduling if both tasks are
 *      ready to run at the time of scheduling.
 *
 * \param[out] id		- pointer to task id to schedule
 * \param[in] args		- pointer to task args to pass into task main
 * \param[in] priority		- priority of scheduling for the task
 * \return
 *	MARS_SUCCESS		- successfully scheduled MARS task for execution
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_PARAMS	- bad task id specified
 * \n	MARS_ERROR_STATE	- task is in an invalid state
 */
int mars_task_schedule(struct mars_task_id *id,
			struct mars_task_args *args,
			uint8_t priority);

/**
 * \ingroup group_mars_task
 * \brief [host] Waits for task completion. (Blocking)
 *
 * This function will block until the scheduled task specified is finished.
 *
 * \param[in] id		- pointer to task id to wait for
 * \return
 *	MARS_SUCCESS		- task execution finished
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_PARAMS	- bad task id specified
 * \n	MARS_ERROR_STATE	- task is in an invalid state
 */
int mars_task_wait(struct mars_task_id *id);

/**
 * \ingroup group_mars_task
 * \brief [host] Waits for a task completion. (Non-Blocking)
 *
 * This function will check whether the scheduled task specified is finished
 * or not and return immediately without blocking.
 *
 * \param[in] id		- pointer to task id to wait for
 * \return
 *	MARS_SUCCESS		- task execution finished
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_PARAMS	- bad task id specified
 * \n	MARS_ERROR_STATE	- task is in an invalid state
 * \n	MARS_ERROR_BUSY		- task has not yet finished execution
 */
int mars_task_try_wait(struct mars_task_id *id);

#if defined(__cplusplus)
}
#endif

#endif
