/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be 
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified 
 *  version of the Library in object code or as an executable provided 
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete 
 *   corresponding machine-readable source code for the modified version 
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer 
 *   for a complete machine-readable copy of the corresponding source 
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#ifndef MARS_TASK_QUEUE_H
#define MARS_TASK_QUEUE_H

/**
 * \file
 * \ingroup group_mars_task_queue
 * \brief [host] MARS Task Queue Flag API
 */

#if defined(__cplusplus)
extern "C" {
#endif

#include "mars/mars_task_queue_types.h"

struct mars_context;

/**
 * \ingroup group_mars_task_queue
 * \brief [host] Initializes a task queue.
 *
 * \param[in] mars		- pointer to initialized MARS context
 * \param[out] queue		- pointer to queue instance to initialize
 * \param[in] buffer		- pointer to data buffer managed by queue
 * \param[in] size		- size of each data entry in data buffer
 * \param[in] depth		- maximum number of data entries in data buffer
 * \param[in] direction		- direction of the event flag
 * \return
 *	MARS_SUCCESS		- successfully initialized queue
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_ALIGN	- instance not aligned properly
 * \n	MARS_ERROR_PARAMS	- size is not multiple of 16 bytes or is greater
 *				  than 16KB maximum
 * \n	MARS_ERROR_PARAMS	- depth exceeds allowed limit
 * \n	MARS_ERROR_PARAMS	- invalid direction specified
 */
int mars_task_queue_initialize(struct mars_context *mars,
				struct mars_task_queue *queue,
				void *buffer,
				uint32_t size,
				uint32_t depth,
				uint8_t direction);

/**
 * \ingroup group_mars_task_queue
 * \brief [host] Returns the number of items in the task queue.
 *
 * \param[in] queue		- pointer to initialized queue instance
 * \param[out] count		- pointer to variable to store return count
 * \return
 *	MARS_SUCCESS		- successfully returned count
 * \n	MARS_ERROR_NULL		- ea is 0
 * \n	MARS_ERROR_ALIGN	- ea not aligned properly
 */
int mars_task_queue_count(struct mars_task_queue *queue, uint32_t *count);

/**
 * \ingroup group_mars_task_queue
 * \brief [host] Clears the items in the task queue.
 *
 * \param[in] queue		- pointer to initialized queue instance
 * \return
 *	MARS_SUCCESS		- successfully cleared queue
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_ALIGN	- instance not aligned properly
 */
int mars_task_queue_clear(struct mars_task_queue *queue);

/**
 * \ingroup group_mars_task_queue
 * \brief [host] Pushes the data specified into the task queue. (Blocking)
 *
 * \param[in] queue		- pointer to initialized queue instance
 * \param[in] data		- address of data to be pushed into queue
 * \return
 *	MARS_SUCCESS		- successfully pushed data into queue
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_ALIGN	- instance not aligned properly
 * \n	MARS_ERROR_STATE	- invalid direction
 */
int mars_task_queue_push(struct mars_task_queue *queue, const void *data);

/**
 * \ingroup group_mars_task_queue
 * \brief [host] Pushes the data specified into the task queue. (Non-Blocking)
 *
 * \param[in] queue		- pointer to initialized queue instance
 * \param[in] data		- address of data to be pushed into queue
 * \return
 *	MARS_SUCCESS		- successfully pushed data into queue
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_ALIGN	- instance not aligned properly
 * \n	MARS_ERROR_BUSY		- queue is full
 * \n	MARS_ERROR_STATE	- invalid direction
 */
int mars_task_queue_try_push(struct mars_task_queue *queue, const void *data);

/**
 * \ingroup group_mars_task_queue
 * \brief [host] Pops data from a task queue. (Blocking)
 *
 * \param[in] queue		- pointer to initialized task instance
 * \param[in] data		- address of data to be popped from queue
 * \return
 *	MARS_SUCCESS		- successfully received event
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_ALIGN	- instance not aligned properly
 * \n	MARS_ERROR_STATE	- invalid direction
 */
int mars_task_queue_pop(struct mars_task_queue *queue, void *data);

/**
 * \ingroup group_mars_task_queue
 * \brief [host] Pops data from a task queue. (Non-Blocking)
 *
 * \param[in] queue		- pointer to initialized task instance
 * \param[in] data		- address of data to be popped from queue
 * \return
 *	MARS_SUCCESS		- successfully received event
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_ALIGN	- instance not aligned properly
 * \n	MARS_ERROR_BUSY		- queue is empty
 * \n	MARS_ERROR_STATE	- invalid direction
 */
int mars_task_queue_try_pop(struct mars_task_queue *queue, void *data);

/**
 * \ingroup group_mars_task_queue
 * \brief [host] Pops data from a task queue without removing it. (Blocking)
 *
 * \param[in] queue		- pointer to initialized task instance
 * \param[in] data		- address of data to be popped from queue
 * \return
 *	MARS_SUCCESS		- successfully received event
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_ALIGN	- instance not aligned properly
 * \n	MARS_ERROR_LIMIT	- exceeded limit of max waiting tasks
 * \n	MARS_ERROR_STATE	- invalid direction
 */
int mars_task_queue_peek(struct mars_task_queue *queue, void *data);

/**
 * \ingroup group_mars_task_queue
 * \brief [host] Pops data from a task queue without removing it. (Non-Blocking)
 *
 * \param[in] queue		- pointer to initialized task instance
 * \param[in] data		- address of data to be popped from queue
 * \return
 *	MARS_SUCCESS		- successfully received event
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_ALIGN	- instance not aligned properly
 * \n	MARS_ERROR_BUSY		- queue is empty
 * \n	MARS_ERROR_STATE	- invalid direction
 */
int mars_task_queue_try_peek(struct mars_task_queue *queue, void *data);

#if defined(__cplusplus)
}
#endif

#endif
