/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be 
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified 
 *  version of the Library in object code or as an executable provided 
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete 
 *   corresponding machine-readable source code for the modified version 
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer 
 *   for a complete machine-readable copy of the corresponding source 
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#ifndef MARS_TIMER_H
#define MARS_TIMER_H

/*
 * \ingroup group_mars_timer
 * \brief [host] MARS Timer API
 */

#if defined(__cplusplus)
extern "C" {
#endif

#include <sys/time.h>
#include "mars/mars_debug.h"

/*
 * \ingroup group_mars_timer
 * \brief MARS timer structure
 *
 * This structure is used for storing timer information.
 * When passed into \ref mars_timer_start() and \ref mars_timer_stop(),
 * the us, ms, and s members will hold the time elapsed between the start
 * and stop calls in microseconds, milliseconds, and seconds, respectively.
 * \note This structure should not be directly modified by the user.
 */
struct mars_timer {
	float us;			/**< microseconds elapsed */
	float ms;			/**< milliseconds elapsed */
	float s;			/**< seconds elapsed */

	unsigned long long start;	/**< internal - start time */
	unsigned long long end;		/**< internal - stop time */
};

/*
 * \ingroup group_mars_timer
 * \brief Starts the timer
 *
 * \param[in] timer	- pointer to the timer instance
 *
 * This should be called in pair with \ref mars_timer_stop.
 * After the timer is stopped, the us, ms, and s members of the timer instance
 * will hold the time elapsed between the start and stop calls in microseconds,
 * milliseconds, and seconds, respectively.
 */
static inline void mars_timer_start(struct mars_timer *timer)
{
	MARS_CHECK(timer);

	struct timeval tv;
	gettimeofday(&tv, NULL);
	timer->start = tv.tv_sec * 1000000 + tv.tv_usec;
	timer->end = timer->start;

	timer->us = 0;
	timer->ms = 0;
	timer->s  = 0;
}

/*
 * \ingroup group_mars_timer
 * \brief Stops the timer
 *
 * \param[in,out] timer	- pointer to the timer instance
 *
 * This should be called after a previous call to \ref mars_timer_start.
 * After the timer is stopped, the us, ms, and s members of the timer instance
 * will hold the time elapsed between the start and stop calls in microseconds,
 * milliseconds, and seconds, respectively.
 */
static inline void mars_timer_stop(struct mars_timer *timer)
{
	MARS_CHECK(timer);

	struct timeval tv;
	gettimeofday(&tv, NULL);
	timer->end = tv.tv_sec * 1000000 + tv.tv_usec;

	timer->us = (float)(timer->end - timer->start);
	timer->ms = timer->us * 1.0e-3;
	timer->s  = timer->us * 1.0e-6;
}

#endif
