/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be 
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified 
 *  version of the Library in object code or as an executable provided 
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete 
 *   corresponding machine-readable source code for the modified version 
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer 
 *   for a complete machine-readable copy of the corresponding source 
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#ifndef MARS_DMA_H
#define MARS_DMA_H

/*
 * \file
 * \ingroup group_mars_dma
 * \brief [MPU] MARS DMA API
 */

#if defined(__cplusplus)
extern "C" {
#endif

#include <spu_mfcio.h>
#include "mars/mars_debug.h"

#define MARS_DMA_TAG		31	/* dma tag reserved for MARS */
#define MARS_DMA_TAG_MAX	31	/* dma tag 0~31 */
#define MARS_DMA_ALIGN_MASK	0xf	/* dma 16 byte alignment mask */
#define MARS_DMA_SIZE_MASK	0xf	/* dma 16 byte size mask */
#define MARS_DMA_SIZE_MAX	16384	/* dma single tranfer size max 16KB */

/* size max 16 KB */
static inline unsigned int _list_init(mfc_list_element_t *list,
				unsigned int size, unsigned int eal)
{
	unsigned int count = 0;

	while (size) {
		unsigned int block_size;
		block_size = (size < MARS_DMA_SIZE_MAX) ?
				size : MARS_DMA_SIZE_MAX;
		list[count].size = block_size;
		list[count].eal = eal;
		size -= block_size;
		eal += block_size;
		count++;
	}

	return (count * sizeof(mfc_list_element_t));
}

static inline void _mars_dma_large_get(void *ls, uint64_t ea,
					uint32_t size, uint32_t tag)
{
	mfc_list_element_t list[16] __attribute__((aligned(8)));
	unsigned int list_size;

	list_size = _list_init(list, size, mfc_ea2l(ea));
	mfc_getl((volatile void *)ls, ea, &list, list_size, tag, 0, 0);
}

static inline void _mars_dma_large_put(const void *ls, uint64_t ea,
					uint32_t size, uint32_t tag)
{
	mfc_list_element_t list[16] __attribute__((aligned(8)));
	unsigned int list_size;

	list_size = _list_init(list, size, mfc_ea2l(ea));
	mfc_putl((volatile void *)ls, ea, &list, list_size, tag, 0, 0);
}

/*
 * \ingroup group_mars_dma
 * \brief [MPU] Waits for a dma transfer to complete.
 *
 * This function will wait until all dma transfer operations currently
 * initialized with the specified tag completes.
 *
 * \param[in] tag	- tag identifier of dma transfer to wait completion of\n
 *			[between 0 and 31]
 * \return
 *	none
 */
static inline void mars_dma_wait(uint32_t tag)
{
#if MARS_DEBUG_ENABLE_DMA_VERBOSE
	MARS_PRINTF("tag=%d\n", tag);
#endif
#if MARS_DEBUG_ENABLE_DMA_ASSERTS
	MARS_ASSERT(tag <= MARS_DMA_TAG_MAX);
#endif
	mfc_write_tag_mask(1 << tag);
	mfc_write_tag_update_all();
	mfc_read_tag_status();
	mfc_sync(tag);
}

/*
 * \ingroup group_mars_dma
 * \brief [MPU] Starts dma transfer from host storage to MPU storage.
 *
 * This function will start a dma transfer to copy the specified number of
 * bytes from host storage to MPU storage. This function should be used
 * when transferring data sizes between 16 bytes and 16 KB. For transfers
 * of sizes smaller than 16 bytes use \ref mars_dma_small_get. For transfers
 * of sizes larger than 16 KB use \ref mars_dma_large_get.
 *
 * \param[in] ls	- address of MPU storage to transfer to
 *			[aligned to a 16 byte address]
 * \param[in] ea	- address of host storage to transfer from
 *			[aligned to a 16 byte address]
 * \param[in] size	- size of bytes to transfer
 *			[multiple of 16 bytes and no larger than 16 KB]
 * \param[in] tag	- tag identifier of dma transfer
 *			[between 0 and 31]
 * \return
 *	none
 */
static inline void mars_dma_get(void *ls, uint64_t ea,
					uint32_t size, uint32_t tag)
{
#if MARS_DEBUG_ENABLE_DMA_VERBOSE
	MARS_PRINTF("ls=%p ea=%#llx size=%d tag=%d\n", ls, ea, size, tag);
#endif
#if MARS_DEBUG_ENABLE_DMA_ASSERTS
	MARS_ASSERT(((uintptr_t)ls & MARS_DMA_ALIGN_MASK) == 0);
	MARS_ASSERT(((uintptr_t)ea & MARS_DMA_ALIGN_MASK) == 0);
	MARS_ASSERT((size & MARS_DMA_SIZE_MASK) == 0);
	MARS_ASSERT(size && size <= MARS_DMA_SIZE_MAX);
	MARS_ASSERT(tag <= MARS_DMA_TAG_MAX);
#endif
	mfc_get((volatile void *)ls, ea, size, tag, 0, 0);
}

/*
 * \ingroup group_mars_dma
 * \brief [MPU] Starts dma get operation and waits for completion.
 *
 * \param[in] ls	- address of MPU storage to transfer to
 *			[aligned to a 16 byte address]
 * \param[in] ea	- address of host storage to transfer from
 *			[aligned to a 16 byte address]
 * \param[in] size	- size of bytes to transfer
 *			[multiple of 16 bytes and no larger than 16 KB]
 * \param[in] tag	- tag identifier of dma transfer
 *			[between 0 and 31]
 * \return
 *	none
 */
static inline void mars_dma_get_and_wait(void *ls, uint64_t ea,
					uint32_t size, uint32_t tag)
{
	mars_dma_get(ls, ea, size, tag);
	mars_dma_wait(tag);
}

/*
 * \ingroup group_mars_dma
 * \brief [MPU] Starts dma transfer from MPU storage to host storage.
 *
 * This function will start a dma transfer to copy the specified number of
 * bytes from MPU storage to host storage. This function should be used
 * when transferring data sizes between 16 bytes and 16 KB. For transfers
 * of sizes smaller than 16 bytes use \ref mars_dma_small_get. For transfers
 * of sizes larger than 16 KB use \ref mars_dma_large_get.
 *
 * \param[in] ls	- address of MPU storage to transfer from
 *			[aligned to a 16 byte address]
 * \param[in] ea	- address of host storage to transfer to
 *			[aligned to a 16 byte address]
 * \param[in] size	- size of bytes to transfer
 *			[multiple of 16 bytes and no larger than 16 KB]
 * \param[in] tag	- tag identifier of dma transfer
 *			[between 0 and 31]
 * \return
 *	none
 */
static inline void mars_dma_put(const void *ls, uint64_t ea,
					uint32_t size, uint32_t tag)
{
#if MARS_DEBUG_ENABLE_DMA_VERBOSE
	MARS_PRINTF("ls=%p ea=%#llx size=%d tag=%d\n", ls, ea, size, tag);
#endif
#if MARS_DEBUG_ENABLE_DMA_ASSERTS
	MARS_ASSERT(((uintptr_t)ls & MARS_DMA_ALIGN_MASK) == 0);
	MARS_ASSERT(((uintptr_t)ea & MARS_DMA_ALIGN_MASK) == 0);
	MARS_ASSERT((size & MARS_DMA_SIZE_MASK) == 0);
	MARS_ASSERT(size && size <= MARS_DMA_SIZE_MAX);
	MARS_ASSERT(tag <= MARS_DMA_TAG_MAX);
#endif
	spu_dsync();
	mfc_put((volatile void *)ls, ea, size, tag, 0, 0);
}

/*
 * \ingroup group_mars_dma
 * \brief [MPU] Starts dma put operation and waits for completion.
 *
 * \param[in] ls	- address of MPU storage to transfer from
 *			[aligned to a 16 byte address]
 * \param[in] ea	- address of host storage to transfer to
 *			[aligned to a 16 byte address]
 * \param[in] size	- size of bytes to transfer
 *			[multiple of 16 bytes and no larger than 16 KB]
 * \param[in] tag	- tag identifier of dma transfer
 *			[between 0 and 31]
 * \return
 *	none
 */
static inline void mars_dma_put_and_wait(const void *ls, uint64_t ea,
					uint32_t size, uint32_t tag)
{
	mars_dma_put(ls, ea, size, tag);
	mars_dma_wait(tag);
}

/*
 * \ingroup group_mars_dma
 * \brief [MPU] Starts small dma transfer from host storage to MPU storage.
 *
 * This function will start a dma transfer to copy the specified number of
 * bytes from host storage to MPU storage. This function should be used
 * when transferring data sizes of 1, 2, 4, or 8 bytes. For transfers
 * of larger sizes use \ref mars_dma_get or \ref mars_dma_large_get.
 *
 * \param[in] ls	- address of MPU storage to transfer to
 *			[lower 4 bits must be same as ea]
 * \param[in] ea	- address of host storage to transfer from
 *			[lower 4 bits must be same as ls]
 * \param[in] size	- size of bytes to transfer
 *			[1, 2, 4, or 8 bytes]
 * \param[in] tag	- tag identifier of dma transfer
 *			[between 0 and 31]
 * \return
 *	none
 */
static inline void mars_dma_small_get(void *ls, uint64_t ea,
					uint32_t size, uint32_t tag)
{
#if MARS_DEBUG_ENABLE_DMA_VERBOSE
	MARS_PRINTF("ls=%p ea=%#llx size=%d tag=%d\n", ls, ea, size, tag);
#endif
#if MARS_DEBUG_ENABLE_DMA_ASSERTS
	MARS_ASSERT(((uintptr_t)ls & MARS_DMA_ALIGN_MASK) == 0);
	MARS_ASSERT(((uintptr_t)ea & MARS_DMA_ALIGN_MASK) == 0);
	MARS_ASSERT(size == 1 || size == 2 || size == 4 || size == 8);
	MARS_ASSERT(tag <= MARS_DMA_TAG_MAX);
#endif
	mfc_get(ls, ea, size, tag, 0, 0);
}

/*
 * \ingroup group_mars_dma
 * \brief [MPU] Starts dma small get operation and waits for completion.
 *
 * \param[in] ls	- address of MPU storage to transfer to
 *			[lower 4 bits must be same as ea]
 * \param[in] ea	- address of host storage to transfer from
 *			[lower 4 bits must be same as ls]
 * \param[in] size	- size of bytes to transfer
 *			[1, 2, 4, or 8 bytes]
 * \param[in] tag	- tag identifier of dma transfer
 *			[between 0 and 31]
 * \return
 *	none
 */
static inline void mars_dma_small_get_and_wait(void *ls, uint64_t ea,
					uint32_t size, uint32_t tag)
{
	mars_dma_small_get(ls, ea, size, tag);
	mars_dma_wait(tag);
}

/*
 * \ingroup group_mars_dma
 * \brief [MPU] Starts small dma transfer from MPU storage to host storage.
 *
 * This function will start a dma transfer to copy the specified number of
 * bytes from MPU storage to host storage. This function should be used
 * when transferring data sizes of 1, 2, 4, or 8 bytes. For transfers
 * of larger sizes use \ref mars_dma_get or \ref mars_dma_large_get.
 *
 * \param[in] ls	- address of MPU storage to transfer from
 *			[aligned to a 16 byte address]
 * \param[in] ea	- address of host storage to transfer to
 *			[aligned to a 16 byte address]
 * \param[in] size	- size of bytes to transfer
 *			[multiple of 16 bytes and no larger than 16 KB]
 * \param[in] tag	- tag identifier of dma transfer
 *			[between 0 and 31]
 * \return
 *	none
 */
static inline void mars_dma_small_put(const void *ls, uint64_t ea,
					uint32_t size, uint32_t tag)
{
#if MARS_DEBUG_ENABLE_DMA_VERBOSE
	MARS_PRINTF("ls=%p ea=%#llx size=%d tag=%d\n", ls, ea, size, tag);
#endif
#if MARS_DEBUG_ENABLE_DMA_ASSERTS
	MARS_ASSERT(((uintptr_t)ls & MARS_DMA_ALIGN_MASK) == 0);
	MARS_ASSERT(((uintptr_t)ea & MARS_DMA_ALIGN_MASK) == 0);
	MARS_ASSERT(size == 1 || size == 2 || size == 4 || size == 8);
	MARS_ASSERT(tag <= MARS_DMA_TAG_MAX);
#endif
	spu_dsync();
	mfc_put((volatile void *)ls, ea, size, tag, 0, 0);
}

/*
 * \ingroup group_mars_dma
 * \brief [MPU] Starts dma small put operation and waits for completion.
 *
 * \param[in] ls	- address of MPU storage to transfer from
 *			[aligned to a 16 byte address]
 * \param[in] ea	- address of host storage to transfer to
 *			[aligned to a 16 byte address]
 * \param[in] size	- size of bytes to transfer
 *			[multiple of 16 bytes and no larger than 16 KB]
 * \param[in] tag	- tag identifier of dma transfer
 *			[between 0 and 31]
 * \return
 *	none
 */
static inline void mars_dma_small_put_and_wait(const void *ls, uint64_t ea,
					uint32_t size, uint32_t tag)
{
	mars_dma_small_put(ls, ea, size, tag);
	mars_dma_wait(tag);
}

/*
 * \ingroup group_mars_dma
 * \brief [MPU] Starts large dma transfer from host storage to MPU storage.
 *
 * This function will start a dma transfer to copy the specified number of
 * bytes from host storage to MPU storage. This function should be used
 * when transferring data sizes of 16 KB and larger. This function can be
 * used for transfer sizes of 16KB and smaller, but for efficiency transfers
 * of smaller sizes should use \ref mars_dma_get or \ref mars_dma_small_get.
 *
 * \param[in] ls	- address of MPU storage to transfer to
 *			[aligned to a 16 byte address]
 * \param[in] ea	- address of host storage to transfer from
 *			[aligned to a 16 byte address]
 * \param[in] size	- size of bytes to transfer
			[16 KB and larger recommended]
 * \param[in] tag	- tag identifier of dma transfer
 *			[between 0 and 31]
 * \return
 *	none
 */
static inline void mars_dma_large_get(void *ls, uint64_t ea,
					uint32_t size, uint32_t tag)
{
#if MARS_DEBUG_ENABLE_DMA_VERBOSE
	MARS_PRINTF("ls=%p ea=%#llx size=%d tag=%d\n", ls, ea, size, tag);
#endif
#if MARS_DEBUG_ENABLE_DMA_ASSERTS
	MARS_ASSERT(((uintptr_t)ls & MARS_DMA_ALIGN_MASK) == 0);
	MARS_ASSERT(((uintptr_t)ea & MARS_DMA_ALIGN_MASK) == 0);
	MARS_ASSERT(size);
	MARS_ASSERT(tag <= MARS_DMA_TAG_MAX);
#endif
	_mars_dma_large_get(ls, ea, size, tag);
}

/*
 * \ingroup group_mars_dma
 * \brief [MPU] Starts dma large get operation and waits for completion.
 *
 * \param[in] ls	- address of MPU storage to transfer to
 *			[aligned to a 16 byte address]
 * \param[in] ea	- address of host storage to transfer from
 *			[aligned to a 16 byte address]
 * \param[in] size	- size of bytes to transfer
			[16 KB and larger recommended]
 * \param[in] tag	- tag identifier of dma transfer
 *			[between 0 and 31]
 * \return
 *	none
 */
static inline void mars_dma_large_get_and_wait(void *ls, uint64_t ea,
						uint32_t size, uint32_t tag)
{
	mars_dma_large_get(ls, ea, size, tag);
	mars_dma_wait(tag);
}

/*
 * \ingroup group_mars_dma
 * \brief [MPU] Starts large dma transfer from MPU storage to host storage.
 *
 * This function will start a dma transfer to copy the specified number of
 * bytes from MPU storage to host storage. This function should be used
 * when transferring data sizes of 16 KB and larger. This function can be
 * used for transfer sizes of 16KB and smaller, but for efficiency transfers
 * of smaller sizes should use \ref mars_dma_get or \ref mars_dma_small_get.
 *
 * \param[in] ls	- address of MPU storage to transfer from
 *			[aligned to a 16 byte address]
 * \param[in] ea	- address of host storage to transfer to
 *			[aligned to a 16 byte address]
 * \param[in] size	- size of bytes to transfer
 *			[multiple of 16 bytes and no larger than 16 KB]
 * \param[in] tag	- tag identifier of dma transfer
 *			[between 0 and 31]
 * \return
 *	none
 */
static inline void mars_dma_large_put(const void *ls, uint64_t ea,
					uint32_t size, uint32_t tag)
{
#if MARS_DEBUG_ENABLE_DMA_VERBOSE
	MARS_PRINTF("ls=%p ea=%#llx size=%d tag=%d\n", ls, ea, size, tag);
#endif
#if MARS_DEBUG_ENABLE_DMA_ASSERTS
	MARS_ASSERT(((uintptr_t)ls & MARS_DMA_ALIGN_MASK) == 0);
	MARS_ASSERT(((uintptr_t)ea & MARS_DMA_ALIGN_MASK) == 0);
	MARS_ASSERT(size);
	MARS_ASSERT(tag <= MARS_DMA_TAG_MAX);
#endif
	_mars_dma_large_put(ls, ea, size, tag);
}

/*
 * \ingroup group_mars_dma
 * \brief [MPU] Starts dma large put operation and waits for completion.
 *
 * \param[in] ls	- address of MPU storage to transfer from
 *			[aligned to a 16 byte address]
 * \param[in] ea	- address of host storage to transfer to
 *			[aligned to a 16 byte address]
 * \param[in] size	- size of bytes to transfer
 *			[multiple of 16 bytes and no larger than 16 KB]
 * \param[in] tag	- tag identifier of dma transfer
 *			[between 0 and 31]
 * \return
 *	none
 */
static inline void mars_dma_large_put_and_wait(const void *ls, uint64_t ea,
					uint32_t size, uint32_t tag)
{
	mars_dma_large_put(ls, ea, size, tag);
	mars_dma_wait(tag);
}

#if defined(__cplusplus)
}
#endif

#endif
