/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be 
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified 
 *  version of the Library in object code or as an executable provided 
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete 
 *   corresponding machine-readable source code for the modified version 
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer 
 *   for a complete machine-readable copy of the corresponding source 
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#ifndef MARS_TASK_H
#define MARS_TASK_H

/**
 * \file
 * \ingroup group_mars_task
 * \brief [MPU] MARS Task API
 */

#if defined(__cplusplus)
extern "C" {
#endif

#include "mars/mars_task_barrier.h"
#include "mars/mars_task_event_flag.h"
#include "mars/mars_task_queue.h"
#include "mars/mars_task_semaphore.h"
#include "mars/mars_task_signal.h"
#include "mars/mars_task_types.h"

/**
 * \ingroup group_mars_task
 * \brief [MPU] Entry point for task.
 *
 * This function is the main entry point for the task program. All task
 * programs will need to have a definition of this function. The arguments
 * passed into this function are specified during task initialization through
 * the call to \ref mars_task_initialize.
 *
 * \param[in] args		- pointer to task args structure in MPU storage
 * \return
 *	user specified
 */
int mars_task_main(const struct mars_task_args *args);

/**
 * \ingroup group_mars_task
 * \brief [MPU] Exits and terminates task.
 *
 * This function causes the task to exit and terminate execution. Calling this
 * function will cause the task to enter the finished state, and will no
 * longer be scheduled to run. This function does not need to be called when
 * returning from \ref mars_task_main since it is called automatically.
 */
void mars_task_exit(void);

/**
 * \ingroup group_mars_task
 * \brief [MPU] Yields caller task so other workloads can run.
 *
 * This function causes the task to yield and allow other tasks to be
 * scheduled to run if available. If other tasks are available to be scheduled
 * to run then this task enters the waiting state until the next time it is
 * scheduled to run. If there are no other tasks available to be scheduled at
 * the time of this function call, then this function has no effect and will
 * return immediately.
 *
 * \return
 *	MARS_SUCCESS		- successfully yielded task
 * \n	MARS_ERROR_FORMAT	- no context save area specified
 */
int mars_task_yield(void);

/**
 * \ingroup group_mars_task
 * \brief [MPU] Schedules a MARS task for execution.
 *
 * This function schedules the task specified for execution.
 * The actual time of execution is determined by the scheduler.
 * Once the task is scheduled for execution by this function, it may not be
 * scheduled for execution until previous execution has finished.
 * You can wait for task completion by calling \ref mars_task_wait or
 * \ref mars_task_try_wait.
 *
 * - \e args: The task args (\ref mars_task_args) are optional, and only need
 *      to be passed in if the task executable expects it. If you specify NULL
 *      the task args passed into the task executable's main function will not
 *      be initialized.
 * \n \n
 * - \e priority: This is the priority of the task between 0 to 255, 0 being
 *      the lowest priority and 255 being the highest priority. Tasks with
 *      higher priority will be prioritized during scheduling if both tasks are
 *      ready to run at the time of scheduling.
 *
 * \param[out] id		- pointer to task id to schedule
 * \param[in] args		- pointer to task args to pass into task main
 * \param[in] priority		- priority of scheduling for the task
 * \return
 *	MARS_SUCCESS		- successfully scheduled MARS task for execution
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_PARAMS	- bad task id specified
 * \n	MARS_ERROR_STATE	- task is in an invalid state
 */
int mars_task_schedule(struct mars_task_id *id,
			struct mars_task_args *args,
			uint8_t priority);

/**
 * \ingroup group_mars_task
 * \brief [MPU] Waits for task completion. (Blocking)
 *
 * This function will block until the scheduled task specified is finished.
 *
 * \param[in] id		- pointer to task id to wait for
 * \return
 *	MARS_SUCCESS		- task execution finished
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_PARAMS	- bad task id specified
 * \n	MARS_ERROR_STATE	- task is in an invalid state
 * \n	MARS_ERROR_FORMAT	- no context save area specified
 */
int mars_task_wait(struct mars_task_id *id);

/**
 * \ingroup group_mars_task
 * \brief [MPU] Waits for a task completion. (Non-Blocking)
 *
 * This function will check whether the scheduled task specified is finished
 * or not and return immediately without blocking.
 *
 * \param[in] id		- pointer to task id to wait for
 * \return
 *	MARS_SUCCESS		- task execution finished
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_PARAMS	- bad task id specified
 * \n	MARS_ERROR_STATE	- task is in an invalid state
 * \n	MARS_ERROR_BUSY		- task has not yet finished execution
 */
int mars_task_try_wait(struct mars_task_id *id);

/**
 * \ingroup group_mars_task
 * \brief [MPU] Gets id of kernel that task is currently being executed on.
 *
 * \return
 *	id of MARS kernel [0 to number of MPUS used by MARS context]
 */
uint32_t mars_task_get_kernel_id(void);

/**
 * \ingroup group_mars_task
 * \brief [MPU] Gets id of caller task.
 *
 * \return
 *	pointer to task id in MPU storage
 */
const struct mars_task_id *mars_task_get_id(void);

/**
 * \ingroup group_mars_task
 * \brief [MPU] Gets name of caller task.
 *
 * \return
 *	pointer to task name in MPU storage
 */
const char *mars_task_get_name(void);

#if defined(__cplusplus)
}
#endif

#endif
