/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be 
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified 
 *  version of the Library in object code or as an executable provided 
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete 
 *   corresponding machine-readable source code for the modified version 
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer 
 *   for a complete machine-readable copy of the corresponding source 
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#ifndef MARS_TASK_EVENT_FLAG_H
#define MARS_TASK_EVENT_FLAG_H

/**
 * \file
 * \ingroup group_mars_task_event_flag
 * \brief [MPU] MARS Task Event Flag API
 */

#if defined(__cplusplus)
extern "C" {
#endif

#include "mars/mars_task_event_flag_types.h"

/**
 * \ingroup group_mars_task_event_flag
 * \brief [MPU] Initializes a task event flag.
 *
 * \param[out] event_flag_ea	- ea of event flag instance to initialize
 * \param[in] direction		- direction of the event flag
 * \param[in] clear_mode	- behavior of how the event flag is cleared
 * \return
 *	MARS_SUCCESS		- successfully initialized event flag
 * \n	MARS_ERROR_NULL		- ea is 0
 * \n	MARS_ERROR_ALIGN	- ea not aligned properly
 * \n	MARS_ERROR_PARAMS	- invalid direction or clear mode specified
 */
int mars_task_event_flag_initialize(uint64_t event_flag_ea,
				uint8_t direction,
				uint8_t clear_mode);

/**
 * \ingroup group_mars_task_event_flag
 * \brief [MPU] Clears the bits specified in the task event flag.
 *
 * \param[in] event_flag_ea	- ea of initialized event flag instance
 * \param[in] bits		- bits to clear
 * \return
 *	MARS_SUCCESS		- successfully cleared event flag bits
 * \n	MARS_ERROR_NULL		- ea is 0\n
 * \n	MARS_ERROR_ALIGN	- ea not aligned properly
 */
int mars_task_event_flag_clear(uint64_t event_flag_ea,
				uint32_t bits);

/**
 * \ingroup group_mars_task_event_flag
 * \brief [MPU] Sets the bits specified in the task event flag.
 *
 * \param[in] event_flag_ea	- ea of initialized event flag instance
 * \param[in] bits		- bits to set
 * \return
 *	MARS_SUCCESS		- successfully cleared event flag bits
 * \n	MARS_ERROR_NULL		- ea is 0
 * \n	MARS_ERROR_ALIGN	- ea not aligned properly
 * \n	MARS_ERROR_STATE	- invalid direction
 */
int mars_task_event_flag_set(uint64_t event_flag_ea,
				uint32_t bits);

/**
 * \ingroup group_mars_task_event_flag
 * \brief [MPU] Waits on a task event flag. (Blocking)
 *
 * \param[in] event_flag_ea	- ea of initialized event flag instance
 * \param[in] mask		- bit mask to test event flag bits against
 * \param[in] mask_mode		- specifies how to mask bits (AND, OR)
 * \return
 *	MARS_SUCCESS		- successfully received event
 * \n	MARS_ERROR_NULL		- ea is 0
 * \n	MARS_ERROR_ALIGN	- ea not aligned properly
 * \n	MARS_ERROR_PARAMS	- invalid mask_mode
 * \n	MARS_ERROR_STATE	- invalid direction
 * \n	MARS_ERROR_LIMIT	- exceeded limit of max waiting tasks
 * \n	MARS_ERROR_FORMAT	- no context save area specified
 */
int mars_task_event_flag_wait(uint64_t event_flag_ea,
				uint32_t mask,
				uint8_t mask_mode);

/**
 * \ingroup group_mars_task_event_flag
 * \brief [MPU] Waits on a task event flag. (Non-Blocking)
 *
 * \param[in] event_flag_ea	- ea of initialized event flag instance
 * \param[in] mask		- bit mask to test event flag bits against
 * \param[in] mask_mode		- specifies how to mask bits (AND, OR)
 * \return
 *	MARS_SUCCESS		- successfully received event
 * \n	MARS_ERROR_NULL		- ea is 0
 * \n	MARS_ERROR_ALIGN	- ea not aligned properly
 * \n	MARS_ERROR_PARAMS	- invalid mask_mode
 * \n	MARS_ERROR_STATE	- invalid direction
 * \n	MARS_ERROR_BUSY		- event not yet received
 */
int mars_task_event_flag_try_wait(uint64_t event_flag_ea,
				uint32_t mask,
				uint8_t mask_mode);

#if defined(__cplusplus)
}
#endif

#endif
