/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be 
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified 
 *  version of the Library in object code or as an executable provided 
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete 
 *   corresponding machine-readable source code for the modified version 
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer 
 *   for a complete machine-readable copy of the corresponding source 
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#ifndef MARS_TASK_QUEUE_H
#define MARS_TASK_QUEUE_H

/**
 * \file
 * \ingroup group_mars_task_queue
 * \brief [MPU] MARS Task Queue API
 */

#if defined(__cplusplus)
extern "C" {
#endif

#include "mars/mars_task_queue_types.h"

/**
 * \ingroup group_mars_task_queue
 * \brief [MPU] Initializes a task queue.
 *
 * \param[out] queue_ea		- ea of queue instance to initialize
 * \param[in] buffer_ea		- ea of data buffer
 * \param[in] size		- size of each data entry in data buffer
 * \param[in] depth		- maximum number of data entries in data buffer
 * \param[in] direction		- direction of the queue
 * \return
 *	MARS_SUCCESS		- successfully initialized queue
 * \n	MARS_ERROR_NULL		- ea is 0
 * \n	MARS_ERROR_ALIGN	- ea not aligned properly
 * \n	MARS_ERROR_PARAMS	- size is not multiple of 16 bytes or is greater
 *				  than 16KB maximum
 * \n	MARS_ERROR_PARAMS	- depth exceeds allowed limit
 * \n	MARS_ERROR_PARAMS	- invalid direction specified
 */
int mars_task_queue_initialize(uint64_t queue_ea,
				uint64_t buffer_ea,
				uint32_t size,
				uint32_t depth,
				uint8_t direction);

/**
 * \ingroup group_mars_task_queue
 * \brief [MPU] Returns the number of items in the task queue.
 *
 * \param[in] queue_ea		- ea of initialized queue instance
 * \param[out] count		- pointer to variable to store return count
 * \return
 *	MARS_SUCCESS		- successfully returned count
 * \n	MARS_ERROR_NULL		- ea is 0
 * \n	MARS_ERROR_ALIGN	- ea not aligned properly
 */
int mars_task_queue_count(uint64_t queue_ea, uint32_t *count);

/**
 * \ingroup group_mars_task_queue
 * \brief [MPU] Clears the items in the task queue.
 *
 * \param[in] queue_ea		- ea of initialized queue instance
 * \return
 *	MARS_SUCCESS		- successfully cleared queue
 * \n	MARS_ERROR_NULL		- ea is 0
 * \n	MARS_ERROR_ALIGN	- ea not aligned properly
 */
int mars_task_queue_clear(uint64_t queue_ea);

/**
 * \ingroup group_mars_task_queue
 * \brief [MPU] Pushes the data specified into the task queue. (Blocking)
 *
 * \param[in] queue_ea		- ea of initialized queue instance
 * \param[in] data		- address of data to be pushed into queue
 * \return
 *	MARS_SUCCESS		- successfully pushed data into queue
 * \n	MARS_ERROR_NULL		- ea is 0
 * \n	MARS_ERROR_ALIGN	- ea not aligned properly
 * \n	MARS_ERROR_LIMIT	- exceeded limit of max waiting tasks
 * \n	MARS_ERROR_STATE	- invalid direction
 */
int mars_task_queue_push(uint64_t queue_ea, const void *data);

/**
 * \ingroup group_mars_task_queue
 * \brief [MPU] Begins push operation on a task queue. (Blocking)
 *
 * \param[in] queue_ea		- ea of initialized queue instance
 * \param[in] data		- address of data to be pushed into queue
 * \param[in] tag		- tag identifier for memory transfer
 * \return
 *	MARS_SUCCESS		- successfully pushed data into queue
 * \n	MARS_ERROR_NULL		- ea is 0
 * \n	MARS_ERROR_ALIGN	- ea not aligned properly
 * \n	MARS_ERROR_LIMIT	- exceeded limit of max waiting tasks
 * \n	MARS_ERROR_STATE	- invalid direction
 * \n	MARS_ERROR_PARAMS	- invalid tag
 */
int mars_task_queue_push_begin(uint64_t queue_ea, const void *data,
				uint32_t tag);

/**
 * \ingroup group_mars_task_queue
 * \brief [MPU] Completes push operation on a task queue.
 *
 * \param[in] queue_ea		- ea of initialized queue instance
 * \param[in] tag		- tag identifier for memory transfer
 * \return
 *	MARS_SUCCESS		- successfully pushed data into queue
 * \n	MARS_ERROR_NULL		- ea is 0
 * \n	MARS_ERROR_ALIGN	- ea not aligned properly
 * \n	MARS_ERROR_STATE	- invalid direction
 * \n	MARS_ERROR_PARAMS	- invalid tag
 */
int mars_task_queue_push_end(uint64_t queue_ea, uint32_t tag);

/**
 * \ingroup group_mars_task_queue
 * \brief [MPU] Pushes the data specified into the task queue. (Non-Blocking)
 *
 * \param[in] queue_ea		- ea of initialized queue instance
 * \param[in] data		- address of data to be pushed into queue
 * \return
 *	MARS_SUCCESS		- successfully pushed data into queue
 * \n	MARS_ERROR_NULL		- ea is 0
 * \n	MARS_ERROR_ALIGN	- ea not aligned properly
 * \n	MARS_ERROR_BUSY		- queue is full
 * \n	MARS_ERROR_STATE	- invalid direction
 */
int mars_task_queue_try_push(uint64_t queue_ea, const void *data);

/**
 * \ingroup group_mars_task_queue
 * \brief [MPU] Begins push operation on a task queue. (Non-Blocking)
 *
 * \param[in] queue_ea		- ea of initialized queue instance
 * \param[in] data		- address of data to be pushed into queue
 * \param[in] tag		- tag identifier for memory transfer
 * \return
 *	MARS_SUCCESS		- successfully pushed data into queue
 * \n	MARS_ERROR_NULL		- ea is 0
 * \n	MARS_ERROR_ALIGN	- ea not aligned properly
 * \n	MARS_ERROR_BUSY		- queue is full
 * \n	MARS_ERROR_STATE	- invalid direction
 * \n	MARS_ERROR_PARAMS	- invalid tag
 */
int mars_task_queue_try_push_begin(uint64_t queue_ea, const void *data,
				uint32_t tag);

/**
 * \ingroup group_mars_task_queue
 * \brief [MPU] Pops data from a task queue. (Blocking)
 *
 * \param[in] queue_ea		- ea of initialized queue instance
 * \param[in] data		- address of data to be popped from queue
 * \return
 *	MARS_SUCCESS		- successfully popped data from queue
 * \n	MARS_ERROR_NULL		- ea or data is 0
 * \n	MARS_ERROR_ALIGN	- ea or data not aligned properly
 * \n	MARS_ERROR_LIMIT	- exceeded limit of max waiting tasks
 * \n	MARS_ERROR_STATE	- invalid direction
 */
int mars_task_queue_pop(uint64_t queue_ea, void *data);

/**
 * \ingroup group_mars_task_queue
 * \brief [MPU] Begins pop operation on a task queue. (Blocking)
 *
 * \param[in] queue_ea		- ea of initialized queue instance
 * \param[in] data		- address of data to be popped from queue
 * \param[in] tag		- tag identifier for memory transfer
 * \return
 *	MARS_SUCCESS		- successfully popped data from queue
 * \n	MARS_ERROR_NULL		- ea or data is 0
 * \n	MARS_ERROR_ALIGN	- ea or data not aligned properly
 * \n	MARS_ERROR_LIMIT	- exceeded limit of max waiting tasks
 * \n	MARS_ERROR_STATE	- invalid direction
 * \n	MARS_ERROR_PARAMS	- invalid tag
 */
int mars_task_queue_pop_begin(uint64_t queue_ea, void *data, uint32_t tag);

/**
 * \ingroup group_mars_task_queue
 * \brief [MPU] Completes pop operation on a task queue.
 *
 * \param[in] queue_ea		- ea of initialized queue instance
 * \param[in] tag		- tag identifier for memory transfer
 * \return
 *	MARS_SUCCESS		- successfully popped data from queue
 * \n	MARS_ERROR_NULL		- ea or data is 0
 * \n	MARS_ERROR_ALIGN	- ea or data not aligned properly
 * \n	MARS_ERROR_STATE	- invalid direction
 * \n	MARS_ERROR_PARAMS	- invalid tag
 */
int mars_task_queue_pop_end(uint64_t queue_ea, uint32_t tag);

/**
 * \ingroup group_mars_task_queue
 * \brief [MPU] Pops data from a task queue. (Non-Blocking)
 *
 * \param[in] queue_ea		- ea of initialized queue instance
 * \param[in] data		- address for data to be popped from queue
 * \return
 *	MARS_SUCCESS		- successfully popped data from queue
 * \n	MARS_ERROR_NULL		- ea or data is 0
 * \n	MARS_ERROR_ALIGN	- ea or data not aligned properly
 * \n	MARS_ERROR_BUSY		- queue is empty
 * \n	MARS_ERROR_STATE	- invalid direction
 */
int mars_task_queue_try_pop(uint64_t queue_ea, void *data);

/**
 * \ingroup group_mars_task_queue
 * \brief [MPU] Begins pop operation on a task queue. (Non-Blocking)
 *
 * \param[in] queue_ea		- ea of initialized queue instance
 * \param[in] data		- address for data to be popped from queue
 * \param[in] tag		- tag identifier for memory transfer
 * \return
 *	MARS_SUCCESS		- successfully popped data from queue
 * \n	MARS_ERROR_NULL		- ea or data is 0
 * \n	MARS_ERROR_ALIGN	- ea or data not aligned properly
 * \n	MARS_ERROR_BUSY		- queue is empty
 * \n	MARS_ERROR_STATE	- invalid direction
 * \n	MARS_ERROR_PARAMS	- invalid tag
 */
int mars_task_queue_try_pop_begin(uint64_t queue_ea, void *data, uint32_t tag);

/**
 * \ingroup group_mars_task_queue
 * \brief [MPU] Pops data from a task queue without removing it. (Blocking)
 *
 * \param[in] queue_ea		- ea of initialized queue instance
 * \param[in] data		- address of data to be popped from queue
 * \return
 *	MARS_SUCCESS		- successfully popped data from queue
 * \n	MARS_ERROR_NULL		- ea or data is 0
 * \n	MARS_ERROR_ALIGN	- ea or data not aligned properly
 * \n	MARS_ERROR_LIMIT	- exceeded limit of max waiting tasks
 * \n	MARS_ERROR_STATE	- invalid direction
 */
int mars_task_queue_peek(uint64_t queue_ea, void *data);

/**
 * \ingroup group_mars_task_queue
 * \brief [MPU] Begins peek operation on a task queue. (Blocking)
 *
 * \param[in] queue_ea		- ea of initialized queue instance
 * \param[in] data		- address of data to be popped from queue
 * \param[in] tag		- tag identifier for memory transfer
 * \return
 *	MARS_SUCCESS		- successfully popped data from queue
 * \n	MARS_ERROR_NULL		- ea or data is 0
 * \n	MARS_ERROR_ALIGN	- ea or data not aligned properly
 * \n	MARS_ERROR_LIMIT	- exceeded limit of max waiting tasks
 * \n	MARS_ERROR_STATE	- invalid direction
 * \n	MARS_ERROR_PARAMS	- invalid tag
 */
int mars_task_queue_peek_begin(uint64_t queue_ea, void *data, uint32_t tag);

/**
 * \ingroup group_mars_task_queue
 * \brief [MPU] Completes peek operation on a task queue.
 *
 * \param[in] queue_ea		- ea of initialized queue instance
 * \param[in] tag		- tag identifier for memory transfer
 * \return
 *	MARS_SUCCESS		- successfully popped data from queue
 * \n	MARS_ERROR_NULL		- ea or data is 0
 * \n	MARS_ERROR_ALIGN	- ea or data not aligned properly
 * \n	MARS_ERROR_STATE	- invalid direction
 * \n	MARS_ERROR_PARAMS	- invalid tag
 */
int mars_task_queue_peek_end(uint64_t queue_ea, uint32_t tag);

/**
 * \ingroup group_mars_task_queue
 * \brief [MPU] Pops data from a task queue without removing it. (Non-Blocking)
 *
 * \param[in] queue_ea		- ea of initialized queue instance
 * \param[in] data		- address of data to be popped from queue
 * \return
 *	MARS_SUCCESS		- successfully popped data from queue
 * \n	MARS_ERROR_NULL		- ea or data is 0
 * \n	MARS_ERROR_ALIGN	- ea or data not aligned properly
 * \n	MARS_ERROR_LIMIT	- exceeded limit of max waiting tasks
 * \n	MARS_ERROR_STATE	- invalid direction
 */
int mars_task_queue_try_peek(uint64_t queue_ea, void *data);

/**
 * \ingroup group_mars_task_queue
 * \brief [MPU] Begins peek operation on a task queue. (Non-Blocking)
 *
 * \param[in] queue_ea		- ea of initialized queue instance
 * \param[in] data		- address of data to be popped from queue
 * \param[in] tag		- tag identifier for memory transfer
 * \return
 *	MARS_SUCCESS		- successfully popped data from queue
 * \n	MARS_ERROR_NULL		- ea or data is 0
 * \n	MARS_ERROR_ALIGN	- ea or data not aligned properly
 * \n	MARS_ERROR_LIMIT	- exceeded limit of max waiting tasks
 * \n	MARS_ERROR_STATE	- invalid direction
 * \n	MARS_ERROR_PARAMS	- invalid tag
 */
int mars_task_queue_try_peek_begin(uint64_t queue_ea, void *data, uint32_t tag);

#if defined(__cplusplus)
}
#endif

#endif
