/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be 
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified 
 *  version of the Library in object code or as an executable provided 
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete 
 *   corresponding machine-readable source code for the modified version 
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer 
 *   for a complete machine-readable copy of the corresponding source 
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#ifndef MARS_TIMER_H
#define MARS_TIMER_H

/*
 * \ingroup group_mars_timer
 * \brief MARS Timer API [MPU]
 */

#if defined(__cplusplus)
extern "C" {
#endif

#include <spu_intrinsics.h>
#include "mars/mars_debug.h"

static const unsigned int mars_timebase = 79800000;

/*
 * \ingroup group_mars_timer
 * \brief MARS timer structure
 *
 * This structure is used for storing timer information.
 * When passed into \ref mars_timer_start() and \ref mars_timer_stop(),
 * the us, ms, and s members will hold the time elapsed between the start
 * and stop calls in microseconds, milliseconds, and seconds, respectively.
 * \note This structure should not be directly modified by the user.
 */
struct mars_timer {
	float us;		/**< microseconds elapsed */
	float ms;		/**< milliseconds elapsed */
	float s;		/**< seconds elapsed */

	unsigned int start;	/**< internal - start time */
	unsigned int end;	/**< internal - stop time */
};

/*
 * \ingroup group_mars_timer
 * \brief Starts the timer
 *
 * This should be called in pair with \ref mars_timer_stop.
 * After the timer is stopped, the us, ms, and s members of the timer instance
 * will hold the time elapsed between the start and stop calls in microseconds,
 * milliseconds, and seconds, respectively.
 *
 * \param[in] timer	- pointer to the timer instance
 */
static inline void mars_timer_start(struct mars_timer *timer)
{
	MARS_CHECK(timer);

	spu_writech(SPU_WrDec, 0xffffffff);
	timer->start = spu_readch(SPU_RdDec);

	timer->us = 0;
	timer->ms = 0;
	timer->s  = 0;
}

/*
 * \ingroup group_mars_timer
 * \brief [MPU] Stops the timer
 *
 * This should be called after a previous call to \ref mars_timer_start.
 * After the timer is stopped, the us, ms, and s members of the timer instance
 * will hold the time elapsed between the start and stop calls in microseconds,
 * milliseconds, and seconds, respectively.
 *
 * \param[in,out] timer	- pointer to the timer instance
 */
static inline void mars_timer_stop(struct mars_timer *timer)
{
	MARS_CHECK(timer);

	timer->end = spu_readch(SPU_RdDec);

	timer->s  = (float)(timer->start - timer->end) / mars_timebase;
	timer->ms = timer->s * 1.0e3;
	timer->us = timer->s * 1.0e6;
}

#endif
