/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be 
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified 
 *  version of the Library in object code or as an executable provided 
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete 
 *   corresponding machine-readable source code for the modified version 
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer 
 *   for a complete machine-readable copy of the corresponding source 
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include <unistd.h>
#include <string.h>

#include "mars/mars_workload_queue.h"
#include "mars/mars_kernel_types.h"
#include "mars/mars_error.h"
#include "mars/mars_debug.h"

int workload_queue_initialize(struct mars_workload_queue *queue)
{
	MARS_CHECK_RET(queue, MARS_ERROR_NULL);

	memset(queue, 0, sizeof(struct mars_workload_queue));

	queue->header.queue_ea = (uint64_t)(uintptr_t)queue;
	queue->header.context_ea = (uint64_t)(uintptr_t)&queue->context;
	queue->header.count = 0;
	queue->header.flag = MARS_FLAG_NONE;

	return MARS_SUCCESS;
}

int workload_queue_finalize(struct mars_workload_queue *queue)
{
	MARS_CHECK_RET(queue, MARS_ERROR_NULL);
	MARS_CHECK_RET(!queue->header.count, MARS_ERROR_STATE);

	queue->header.flag = MARS_FLAG_EXIT;

	return MARS_SUCCESS;
}

int workload_queue_add(struct mars_workload_queue *queue, uint16_t *id,
			struct mars_workload_context *workload)
{
	MARS_CHECK_RET(queue, MARS_ERROR_NULL);
	MARS_CHECK_RET(workload, MARS_ERROR_NULL);
	MARS_CHECK_RET(id, MARS_ERROR_NULL);
	MARS_CHECK_RET(workload->type == MARS_WORKLOAD_TYPE_TASK,
			MARS_ERROR_PARAMS);
	MARS_CHECK_RET(queue->header.count < MARS_WORKLOAD_MAX,
			MARS_ERROR_LIMIT);

	int block = 0;
	int index = 0;

	/* search through workload queue for open slot */
	if (queue->header.count) {
		int i;
		int j;

		for (i = 0; i < MARS_WORKLOAD_NUM_BLOCKS; i++) {
			for (j = 0; j < MARS_WORKLOAD_PER_BLOCK; j++) {
				if (queue->block[i].state[j] ==
					MARS_WORKLOAD_STATE_NONE) {
					block = i;
					index = j;
					i = MARS_WORKLOAD_NUM_BLOCKS;
					j = MARS_WORKLOAD_PER_BLOCK;
				}
			}
		}
	}

	/* set return id of added workload context */
	*id = MARS_WORKLOAD_PER_BLOCK * block + index;

	/* copy workload context into workload queue */
	memcpy(&queue->context[*id], workload,
		sizeof(struct mars_workload_context));

	/* update workload queue header info */
	queue->block[block].state[index] = MARS_WORKLOAD_STATE_FINISHED;
	queue->block[block].priority[index] = 0;
	queue->block[block].counter[index] = 0;
	queue->block[block].signal[index] = 0;
	queue->block[block].wait[index] = MARS_WORKLOAD_ID_NONE;
	queue->header.count++;

	return MARS_SUCCESS;
}

int workload_queue_remove(struct mars_workload_queue *queue, uint16_t id,
			struct mars_workload_context *workload)
{
	MARS_CHECK_RET(queue, MARS_ERROR_NULL);
	MARS_CHECK_RET(workload, MARS_ERROR_NULL);
	MARS_CHECK_RET(id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);

	int block = id / MARS_WORKLOAD_PER_BLOCK;
	int index = id % MARS_WORKLOAD_PER_BLOCK;

	MARS_CHECK_RET(queue->block[block].state[index] ==
			MARS_WORKLOAD_STATE_FINISHED, MARS_ERROR_STATE);

	/* copy workload context out from workload queue */
	memcpy(workload, &queue->context[id],
		sizeof(struct mars_workload_context));

	/* update workload queue info */
	queue->block[block].state[index] = MARS_WORKLOAD_STATE_NONE;
	queue->block[block].priority[index] = MARS_WORKLOAD_PRIORITY_MIN;
	queue->block[block].counter[index] = 0;
	queue->block[block].signal[index] = 0;
	queue->block[block].wait[index] = MARS_WORKLOAD_ID_NONE;
	queue->header.count--;

	return MARS_SUCCESS;
}

int workload_queue_context(struct mars_workload_queue *queue, uint16_t id,
			struct mars_workload_context **workload)
{
	MARS_CHECK_RET(queue, MARS_ERROR_NULL);
	MARS_CHECK_RET(workload, MARS_ERROR_NULL);
	MARS_CHECK_RET(id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);

	int block = id / MARS_WORKLOAD_PER_BLOCK;
	int index = id % MARS_WORKLOAD_PER_BLOCK;

	MARS_CHECK_RET(queue->block[block].state[index] !=
			MARS_WORKLOAD_STATE_NONE, MARS_ERROR_STATE);

	*workload = &queue->context[id];

	return MARS_SUCCESS;
}

int workload_queue_schedule(struct mars_workload_queue *queue, uint16_t id,
			uint8_t priority)
{
	MARS_CHECK_RET(queue, MARS_ERROR_NULL);
	MARS_CHECK_RET(id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);

	int block = id / MARS_WORKLOAD_PER_BLOCK;
	int index = id % MARS_WORKLOAD_PER_BLOCK;

	MARS_CHECK_RET(queue->block[block].state[index] ==
			MARS_WORKLOAD_STATE_FINISHED, MARS_ERROR_STATE);

	queue->block[block].state[index] = MARS_WORKLOAD_STATE_READY;
	queue->block[block].priority[index] = priority;
	queue->block[block].counter[index] = 0;
	queue->block[block].signal[index] = 0;

	return MARS_SUCCESS;
}

int workload_queue_wait(struct mars_workload_queue *queue, uint16_t id)
{
	MARS_CHECK_RET(queue, MARS_ERROR_NULL);
	MARS_CHECK_RET(id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);

	int block = id / MARS_WORKLOAD_PER_BLOCK;
	int index = id % MARS_WORKLOAD_PER_BLOCK;

	MARS_CHECK_RET(queue->block[block].state[index] !=
			MARS_WORKLOAD_STATE_NONE, MARS_ERROR_STATE);

	while (*(volatile uint8_t *)&queue->block[block].state[index] !=
		MARS_WORKLOAD_STATE_FINISHED) {
	}

	return MARS_SUCCESS;
}

int workload_queue_try_wait(struct mars_workload_queue *queue, uint16_t id)
{
	MARS_CHECK_RET(queue, MARS_ERROR_NULL);
	MARS_CHECK_RET(id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);

	int block = id / MARS_WORKLOAD_PER_BLOCK;
	int index = id % MARS_WORKLOAD_PER_BLOCK;

	MARS_CHECK_RET(queue->block[block].state[index] !=
			MARS_WORKLOAD_STATE_NONE, MARS_ERROR_STATE);

	if (queue->block[block].state[index] != MARS_WORKLOAD_STATE_FINISHED)
		return MARS_ERROR_BUSY;

	return MARS_SUCCESS;
}

int workload_queue_signal_send(struct mars_workload_queue *queue, uint16_t id)
{
	MARS_CHECK_RET(queue, MARS_ERROR_NULL);
	MARS_CHECK_RET(id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);

	int block = id / MARS_WORKLOAD_PER_BLOCK;
	int index = id % MARS_WORKLOAD_PER_BLOCK;

	MARS_CHECK_RET(queue->block[block].state[index] !=
			MARS_WORKLOAD_STATE_NONE, MARS_ERROR_STATE);

	queue->block[block].signal[index] = 1;

	return MARS_SUCCESS;
}
