/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be 
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified 
 *  version of the Library in object code or as an executable provided 
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete 
 *   corresponding machine-readable source code for the modified version 
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer 
 *   for a complete machine-readable copy of the corresponding source 
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include "mars/mars_kernel.h"

struct mars_kernel_syscalls kernel_syscalls;

void *kernel_stack;
void *process_stack;

void mars_reentry(void)
{
	syscall_resume();
}

void syscall_setup(void)
{
	kernel_syscalls.get_mars_context_ea	= syscall_get_mars_context_ea;
	kernel_syscalls.get_kernel_id		= syscall_get_kernel_id;
	kernel_syscalls.get_workload_id		= syscall_get_workload_id;
	kernel_syscalls.get_workload_ea		= syscall_get_workload_ea;
	kernel_syscalls.get_workload_context	= syscall_get_workload_context;
	kernel_syscalls.init			= syscall_init;
	kernel_syscalls.exit			= syscall_exit;
	kernel_syscalls.yield			= syscall_yield;
	kernel_syscalls.schedule		= syscall_schedule;
	kernel_syscalls.wait			= syscall_wait;
	kernel_syscalls.try_wait		= syscall_try_wait;
	kernel_syscalls.signal_send		= syscall_signal_send;
	kernel_syscalls.signal_wait		= syscall_signal_wait;
	kernel_syscalls.signal_try_wait		= syscall_signal_try_wait;
}

uint64_t syscall_get_mars_context_ea(void)
{
	return kernel_params.mars_context_ea;
}

uint32_t syscall_get_kernel_id(void)
{
	return kernel_params.id;
}

uint16_t syscall_get_workload_id(void)
{
	return workload_get_id();
}

uint64_t syscall_get_workload_ea(void)
{
	return workload_get_ea();
}

struct mars_workload_context *syscall_get_workload_context(void)
{
	return workload_get_context();
}

void syscall_init(void)
{
	/* save the kernel stack
	**
	** We need to offset the stack pointer we want to save so that we
	** return back to the caller of mars_entry and not the caller of
	** the syscall mars_init and when mars_exit is called and pops the
	** stack. But we don't want to modify the stack pointer directly
	** just yet so we can return back to the current caller (mars_entry)
	** so use a volatile register to do the offset.
	** NO FUNCTION CALLS OR STACK PUSHES ALLOWED IN HERE.
	*/
	asm volatile (
		"ai	$75, $sp, 32;"
		"stqa	$75, kernel_stack;"
	);
}

void syscall_exit(void)
{
	/* restore kernel stack */
	asm volatile (
		"lqa	$sp, kernel_stack;"
	);

	/* exit current process */
	workload_stop();
}

void syscall_yield(void)
{
	/* save process stack and restore kernel stack */
	asm volatile (
		"stqa	$sp, process_stack;"
		"lqa	$sp, kernel_stack;"
	);

	/* save current process */
	workload_save();
}

void syscall_resume(void)
{
	/* restore current process */
	workload_restore();

	/* save kernel stack and restore process stack */
	asm volatile (
		"stqa	$sp, kernel_stack;"
		"lqa	$sp, process_stack;"
	);
}

int syscall_schedule(uint16_t workload_id, void *args, uint8_t priority)
{
	return workload_schedule(workload_id, args, priority);
}

int syscall_wait(uint16_t workload_id)
{
	return workload_wait(workload_id);
}

int syscall_try_wait(uint16_t workload_id)
{
	return workload_try_wait(workload_id);
}

int syscall_signal_send(uint16_t workload_id)
{
	return workload_signal_send(workload_id);
}

void syscall_signal_wait(void)
{
	/* save process stack and restore kernel stack */
	asm volatile (
		"stqa	$sp, process_stack;"
		"lqa	$sp, kernel_stack;"
	);

	workload_signal_wait();
}

int syscall_signal_try_wait(void)
{
	return workload_signal_try_wait();
}
