/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be 
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified 
 *  version of the Library in object code or as an executable provided 
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete 
 *   corresponding machine-readable source code for the modified version 
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer 
 *   for a complete machine-readable copy of the corresponding source 
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include "mars/mars_kernel.h"

extern struct mars_workload_queue_header queue_header;
extern struct mars_workload_queue_block queue_block;
extern struct mars_workload_context workload;
extern uint8_t workload_state;
extern uint16_t workload_index;
extern uint64_t workload_ea;

uint16_t workload_get_id(void)
{
	return workload_index;
}

uint64_t workload_get_ea(void)
{
	return workload_ea;
}

struct mars_workload_context *workload_get_context(void)
{
	return &workload;
}

void workload_run(void)
{
	/* set current workload state to running */
	workload_state = MARS_WORKLOAD_STATE_RUNNING;

	/* workload type specific handling */
	switch(workload.type) {
	case MARS_WORKLOAD_TYPE_TASK:
		task_run(&workload.type_context.task);
		break;
	}
}

void workload_stop(void)
{
	/* set current workload state to finished */
	workload_state = MARS_WORKLOAD_STATE_FINISHED;

	/* workload type specific handling */
	switch(workload.type) {
	case MARS_WORKLOAD_TYPE_TASK:
		task_stop(&workload.type_context.task);
		break;
	}
}

void workload_save(void)
{
	/* set current workload state to ready */
	workload_state = MARS_WORKLOAD_STATE_READY;

	/* workload type specific handling */
	switch(workload.type) {
	case MARS_WORKLOAD_TYPE_TASK:
		/* save non-volatile registers */
		registers_save();

		/* save the task context to main memory */
		task_save(&workload.type_context.task);
		break;
	}
}

void workload_restore(void)
{
	/* set current workload state to running */
	workload_state = MARS_WORKLOAD_STATE_RUNNING;

	/* workload type specific handling */
	switch(workload.type) {
	case MARS_WORKLOAD_TYPE_TASK:
		/* restore the task context from main memory */
		task_restore(&workload.type_context.task);

		/* restore non-volatile registers */
		registers_restore();
		break;
	}
}

int workload_schedule(uint16_t workload_id, void *args, uint8_t priority)
{
	MARS_CHECK_RET(workload_id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);
	MARS_CHECK_RET(workload_id != workload_index, MARS_ERROR_PARAMS);

	struct mars_workload_context schedule_workload;
	uint64_t schedule_workload_ea;

	int block = workload_id / MARS_WORKLOAD_PER_BLOCK;
	int index = workload_id % MARS_WORKLOAD_PER_BLOCK;
	int status;

	atomic_event_setup();

	/* attempt to schedule workload until successful */
	do {
		/* reserve cache line of workload queue block */
		atomic_get(&queue_block,
			queue_header.queue_ea +
			offsetof(struct mars_workload_queue, block) +
			sizeof(struct mars_workload_queue_block) * block);

		/* make sure workload is in the correct state */
		if (queue_block.state[index] != MARS_WORKLOAD_STATE_FINISHED) {
			atomic_event_restore();
			return MARS_ERROR_STATE;
		}

		/* calculate context ea from workload id */
		schedule_workload_ea = queue_header.context_ea +
			workload_id * sizeof(struct mars_workload_context);

		/* dma the workload context code into LS from main memory */
		mars_dma_get_and_wait((void *)&schedule_workload,
					schedule_workload_ea,
					sizeof(struct mars_workload_context),
					MARS_DMA_TAG);

		/* workload type specific handling */
		switch(schedule_workload.type) {
		case MARS_WORKLOAD_TYPE_TASK:
			task_schedule(&schedule_workload.type_context.task,
					(struct mars_task_args *)args);
			break;
		}

		/* dma the workload context code into main memory from LS */
		mars_dma_put_and_wait((void *)&schedule_workload,
					schedule_workload_ea,
					sizeof(struct mars_workload_context),
					MARS_DMA_TAG);

		queue_block.state[index] = MARS_WORKLOAD_STATE_READY;
		queue_block.priority[index] = priority;
		queue_block.counter[index] = 0;
		queue_block.signal[index] = 0;
		queue_block.wait[index] = MARS_WORKLOAD_ID_NONE;

		/* attempt to write back workload queue block to cache line */
		status = atomic_put(&queue_block,
			queue_header.queue_ea +
			offsetof(struct mars_workload_queue, block) +
			sizeof(struct mars_workload_queue_block) * block);

		/* write back failed so wait until reservation is lost */
		if (status)
			atomic_event_wait();
	} while (status);

	atomic_event_restore();

	return MARS_SUCCESS;
}

int workload_wait(uint16_t workload_id)
{
	MARS_CHECK_RET(workload_id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);
	MARS_CHECK_RET(workload_id != workload_index, MARS_ERROR_PARAMS);

	int block = workload_index / MARS_WORKLOAD_PER_BLOCK;
	int index = workload_index % MARS_WORKLOAD_PER_BLOCK;
	int status;

	atomic_event_setup();

	/* attempt to update workload queue block until successful */
	do {
		atomic_get(&queue_block,
			queue_header.queue_ea +
			offsetof(struct mars_workload_queue, block) +
			sizeof(struct mars_workload_queue_block) * block);

		/* make sure workload is initialized */
		if (queue_block.state[index] == MARS_WORKLOAD_STATE_NONE) {
			atomic_event_restore();
			return MARS_ERROR_STATE;
		}

		/* set the workload id to wait for */
		queue_block.wait[index] = workload_id;

		/* attempt to write back workload queue block to cache line */
		status = atomic_put(&queue_block,
			queue_header.queue_ea +
			offsetof(struct mars_workload_queue, block) +
			sizeof(struct mars_workload_queue_block) * block);

		/* write back failed so wait until reservation is lost */
		if (status)
			atomic_event_wait();
	} while (status);

	atomic_event_restore();

	return MARS_SUCCESS;
}

int workload_try_wait(uint16_t workload_id)
{
	MARS_CHECK_RET(workload_id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);
	MARS_CHECK_RET(workload_id != workload_index, MARS_ERROR_PARAMS);

	int block = workload_id / MARS_WORKLOAD_PER_BLOCK;
	int index = workload_id % MARS_WORKLOAD_PER_BLOCK;

	atomic_get(&queue_block,
		queue_header.queue_ea +
		offsetof(struct mars_workload_queue, block) +
		sizeof(struct mars_workload_queue_block) * block);

	MARS_CHECK_RET(queue_block.state[index] != MARS_WORKLOAD_STATE_NONE,
			MARS_ERROR_STATE);

	if (queue_block.state[index] != MARS_WORKLOAD_STATE_FINISHED)
		return MARS_ERROR_BUSY;

	return MARS_SUCCESS;
}

int workload_signal_send(uint16_t workload_id)
{
	MARS_CHECK_RET(workload_id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);
	MARS_CHECK_RET(workload_id != workload_index, MARS_ERROR_PARAMS);

	int block = workload_id / MARS_WORKLOAD_PER_BLOCK;
	int index = workload_id % MARS_WORKLOAD_PER_BLOCK;
	int status;

	atomic_event_setup();

	/* attempt to update workload queue block until successful */
	do {
		/* reserve cache line of workload queue block */
		atomic_get(&queue_block,
			queue_header.queue_ea +
			offsetof(struct mars_workload_queue, block) +
			sizeof(struct mars_workload_queue_block) * block);

		/* make sure workload is initialized */
		if (queue_block.state[index] == MARS_WORKLOAD_STATE_NONE) {
			atomic_event_restore();
			return MARS_ERROR_STATE;
		}

		/* set the workload signal */
		queue_block.signal[index] = 1;

		/* attempt to write back workload queue block to cache line */
		status = atomic_put(&queue_block,
			queue_header.queue_ea +
			offsetof(struct mars_workload_queue, block) +
			sizeof(struct mars_workload_queue_block) * block);

		/* write back failed so wait until reservation is lost */
		if (status)
			atomic_event_wait();
	} while (status);

	atomic_event_restore();

	return MARS_SUCCESS;
}

void workload_signal_wait(void)
{
	/* set current workload state to waiting */
	workload_state = MARS_WORKLOAD_STATE_WAITING;

	/* workload type specific handling */
	switch(workload.type) {
	case MARS_WORKLOAD_TYPE_TASK:
		/* save non-volatile registers */
		registers_save();

		/* save the task context to main memory */
		task_save(&workload.type_context.task);
		break;
	}
}

int workload_signal_try_wait(void)
{
	int block = workload_index / MARS_WORKLOAD_PER_BLOCK;
	int index = workload_index % MARS_WORKLOAD_PER_BLOCK;

	atomic_get(&queue_block,
		queue_header.queue_ea +
		offsetof(struct mars_workload_queue, block) +
		sizeof(struct mars_workload_queue_block) * block);

	MARS_CHECK_RET(queue_block.state[index] != MARS_WORKLOAD_STATE_NONE,
			MARS_ERROR_STATE);

	/* return busy if task has not received signal */
	if (!queue_block.signal[index])
		return MARS_ERROR_BUSY;

	return MARS_SUCCESS;
}
